"""
Module for interacting with tj. Mostly dumping tasks in tj3 file.
"""

import dataclasses
import re
from dataclasses import dataclass
from typing import List, Optional, Tuple

from pydash import py_

import jaglar.ganttpro as ganttpro
from jaglar.types import Resource, Task


def normalize_name(name: str) -> str:
    """
    Standardize name string to become a valid tj id
    """

    if not name:
        return name

    name = re.sub(r"[^a-zA-z0-9_]", "_", name.lower())

    if name[0].isdigit():
        name = "a_" + name

    return name


def normalize_resource(resource: Resource) -> Resource:
    return dataclasses.replace(resource, name=normalize_name(resource.name))


def normalize_task(task: Task) -> Task:
    new_task = dataclasses.replace(
        task,
        name=normalize_name(task.name),
        assignee=[normalize_resource(res) for res in task.assignee],
    )

    if task.depends_on:
        new_task.depends_on = [normalize_task(t) for t in task.depends_on]

    return new_task


def namespace_task(task: Task, prefix: str) -> Task:
    new_task = dataclasses.replace(task, name=prefix + "_" + task.name)

    if task.depends_on:
        new_task.depends_on = [namespace_task(t, prefix) for t in task.depends_on]

    return new_task


@dataclass
class Node:
    """
    A node is a piece that maps to a valid taskjuggler code block. This
    doesn't cover macros.
    """

    type: str
    props: List[str]
    children: Optional[List["Node"]] = None


def format_node(node: Node, indent=0, force_brackets=False) -> str:
    """
    Convert node to str that goes in a tj file.
    """

    indent_str = "  " * indent
    top_str = indent_str + node.type + " " + " ".join(node.props)

    if not force_brackets:
        if not node.children:
            return top_str

    blocks = [top_str + " {"]

    for child in (node.children or []):
        blocks.append(format_node(child, indent + 1))

    blocks.append(indent_str + "}")

    return "\n".join(blocks)


def task_to_node(task: Task) -> Node:
    """
    Make node for given task. Note that the effort is considered in
    person-hours.
    """

    dependencies = [
        Node(type="depends", props=[dep.name]) for dep in (task.depends_on or [])
    ]

    assignees = [
        Node(type="allocate", props=[assignee.name]) for assignee in task.assignee
    ]

    return Node(
        type="task",
        props=[task.name, f"\"{task.name}\""],
        children=[
            Node(
                type="effort",
                props=[f"{task.effort}h"]
            ),
            *assignees, *dependencies
        ]
    )


def resource_to_node(resource: Resource) -> Node:
    """
    Convert a resource to tj node.
    """

    return Node(
        type="resource",
        props=[resource.name, f"\"{resource.name}\""],
        children=[
            Node(
                type="limits",
                props=[],
                children=[Node(type="dailymax", props=[f"{resource.hours_per_day}h"])]
            )
        ]
    )


def make_project_node(project_id: str, project_name: str,
                      start_date: str, end_date: str) -> Node:
    """
    Make tj project in a node format. `end_date` can be in the diff format.
    """

    return Node(
        type="project",
        props=[project_id, f"\"{project_name}\"", start_date, end_date]
    )


def make_gantt_node() -> Node:
    return Node(
        type="taskreport",
        props=["\"Gantt Chart\""],
        children=[
            Node(type="formats", props=["html"]),
            Node(type="headline", props=["\"Project Gantt Chart\""]),
            Node(type="columns", props=["hierarchindex, name, start, end, effort, duration, chart"]),
            Node(type="timeformat", props=["\"%a %Y-%m-%d\""])
        ]
    )


def normalize_gantt_pro_projects(*projects: List[ganttpro.Project]) -> Tuple[List[Resource], List[Task]]:
    """
    Convert ganttpro projects to list of valid resources and tasks for tj.
    """

    all_resources = py_.flatten([
        [normalize_resource(res) for res in project.resources]
        for project in projects
    ])
    resources = py_.uniq_by(all_resources, lambda res: res.name)

    tasks = []

    for project in projects:
        prefix = normalize_name(project.name)
        for task in project.tasks:
            tasks.append(namespace_task(normalize_task(task), prefix))

    return resources, tasks


def format_project(project: Node, resources: List[Node], tasks: List[Node], generate_report=True):
    """
    Format the project in tj format. All tasks are assumed to be the lowest
    level one right now since we are mostly interesting in estimating
    timelines.
    """

    formatted_nodes = [
        format_node(project, force_brackets=True),
        *[format_node(res) for res in resources],
        *[format_node(task) for task in tasks]
    ]

    if generate_report:
        formatted_nodes.append(format_node(make_gantt_node()))

    return "\n\n".join(formatted_nodes)



from dataclasses import dataclass
from typing import List, Optional


@dataclass
class Resource:
    name: str
    hours_per_day: int = 8


@dataclass
class Task:
    name: str
    assignee: List[Resource]
    effort: float
    depends_on: Optional[List["Task"]] = None



"""
Module for working with ganttpro exported data.
"""

from typing import Dict, List
from xml.etree.ElementTree import fromstring

import pandas as pd
from pydash import py_
from xmljson import badgerfish as bf

from jaglar.types import Resource, Task
from jaglar.utils import map_keys


def read_xlsx_export(file_path: str):
    return pd.read_excel(file_path, header=3)


def read_xml_export(file_path: str):
    with open(file_path) as fp:
        data = bf.data(fromstring(fp.read()))

    def _clean_key(k: str) -> str:
        try:
            return k.split("}", 1)[1]
        except IndexError:
            return k

    return map_keys(data, _clean_key)


class Project:
    """
    A gantt pro project. This needs both the excel and xml export because both
    have different data formats.
    """

    def __init__(self, name: str, xlsx_path: str, xml_path: str):
        self.name = name
        self.xlsx_data = read_xlsx_export(xlsx_path)
        self.xml_data = read_xml_export(xml_path)

    @property
    def resources(self) -> List[Resource]:
        """
        Parse list of resources using excel data.
        """

        assignments = [
            [name.strip() for name in it.split(",")]
            for it in self.xlsx_data["Assigned to"].dropna().tolist()
        ]

        resources = [Resource(name=name) for name in py_.uniq(py_.flatten(assignments))]

        # We also add a ghost for unassigned items
        resources.append(Resource(name="Ghost"))
        return resources

    @property
    def tasks(self):
        """
        Return tasks with dependencies and resources assigned. This merges
        information from both sources. Also we only consider tasks with type
        task and not groups.
        """

        # NOTE: We use WBS Number for identifying tasks
        excel_info: Dict[str, Task] = {}

        for _, row in self.xlsx_data.iterrows():
            if row["Type"] != "task":
                continue

            try:
                assignees = [Resource(name=name.strip()) for name in row["Assigned to"].split(",")]
            except AttributeError:
                # There are cases where resources are not assigned
                assignees = [Resource(name="Ghost")]

            excel_info[row["WBS Number"]] = Task(
                name=row["Task name / Title"] + " " + row["WBS Number"],
                assignee=assignees,
                effort=int(row["Duration (hours)"])
            )

        # NOTE: There are a few assumptions here about number of projects and
        #       all. This is not foolproof.

        # These tasks are the superset of what we are interested in excel_info
        xml_tasks = self.xml_data["Project"]["Tasks"]["Task"]

        def _sanitize_wbs(wbs: str) -> str:
            wbs = str(wbs)
            # HACK: xml data has extra thing added in WBS
            if "." in wbs:
                return wbs.split(".", 1)[1]
            else:
                return wbs

        uid_to_wbs = {t["UID"]["$"]: _sanitize_wbs(t["WBS"]["$"]) for t in xml_tasks}

        # Mostly mapping from WBS to list of dependencies, also specified in WBS
        xml_info: Dict[str, List[str]] = {}

        for xml_task in xml_tasks:
            wbs = _sanitize_wbs(xml_task["WBS"]["$"])

            if "PredecessorLink" in xml_task:
                pred_link = xml_task["PredecessorLink"]

                if isinstance(pred_link, dict):
                    predecessor_wbses = [uid_to_wbs[pred_link["PredecessorUID"]["$"]]]
                elif isinstance(pred_link, list):
                    predecessor_wbses = [
                        uid_to_wbs[it["PredecessorUID"]["$"]]
                        for it in pred_link
                    ]
                else:
                    raise TypeError(f"Wrong type of PredecessorLink")
                xml_info[wbs] = predecessor_wbses

        # Now we patch the excel_info tasks with dependency information
        for wbs, predecessor_wbses in xml_info.items():
            if wbs not in excel_info:
                continue

            excel_info[wbs].depends_on = [excel_info[p_wbs] for p_wbs in predecessor_wbses]

        return list(excel_info.values())



from typing import Dict


def map_keys(d: Dict, fn) -> Dict:
    """
    Map fn on d's keys
    """

    new_d = {}
    for k, v in d.items():
        if isinstance(v, dict):
            new_d[fn(k)] = map_keys(v, fn)
        elif isinstance(v, list):
            new_d[fn(k)] = [map_keys(it, fn) for it in v]
        else:
            new_d[fn(k)] = v

    return new_d



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jaglar']

package_data = \
{'': ['*']}

install_requires = \
['docopt>=0.6.2,<0.7.0',
 'pandas>=1.0.3,<2.0.0',
 'pydash>=4.7.6,<5.0.0',
 'xlrd>=1.2.0,<2.0.0',
 'xmljson>=0.2.0,<0.3.0']

entry_points = \
{'console_scripts': ['jaglar = jaglar.cli:main']}

setup_kwargs = {
    'name': 'jaglar',
    'version': '0.1.0',
    'description': 'Tool for project management backed by taskjuggler.',
    'long_description': None,
    'author': 'Abhinav Tushar',
    'author_email': 'abhinav@lepisma.xyz',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Vernacular-ai/jaglar/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
